from cyxtrax.simulation import CylindricalProjection, load_mesh, SAVEMODES, utility
from cyxtrax.io import load_atlas
from cyxtrax.mapping import map_geometry_2_projection, map_source_2_cylinder
from pathlib import Path
from jax import numpy as jnp
from jax.scipy.spatial.transform import Rotation
from matplotlib import pyplot as plt, figure


# Just get the temp folder to store the atlas
FOLDER = Path("/mnt/data/xray_studies_data/")
TEMP_FOLDER = FOLDER / "dataset_64"
TEMP_FOLDER = Path('temp')


def main():
    files = list(TEMP_FOLDER.glob("*.h5"))
    if not files:
        raise FileNotFoundError("No .h5 files in temp folder. Run Example 02!")

    maps, points, mesh_object_list = load_atlas(files[-1])
    cyl_proj = CylindricalProjection()

    load_mesh(cyl_proj, mesh_object_list, False)
    print(f"Cone Beam Mode set: {cyl_proj.cone_mode}")

    rot = 30
    x, y = 500 * jnp.cos(rot * jnp.pi / 180), 500 * jnp.sin(rot * jnp.pi / 180)
    source_position = jnp.array([x, y, 0.0]).reshape((1, 3))
    detector_position = jnp.array([-x, -y, 0.0]).reshape((1, 3)) / 2.
    detctor_orientation_scipy = Rotation.from_euler("xyz", [0, -90, rot], degrees=True)

    temp_projection = Path("temp.tif")
    detctor_orientation = detctor_orientation_scipy.as_quat().reshape((1, -1))

    cyl_proj.api.translate(
        "S", source_position[0, 0], source_position[0, 1], source_position[0, 2]
    )
    cyl_proj.api.translate(
        "D", detector_position[0, 0], detector_position[0, 1], detector_position[0, 2]
    )
    cyl_proj.api.rotate_from_quat("D", detctor_orientation[0])
    cyl_proj.api.save_image(temp_projection, save_mode=SAVEMODES.FLOAT_TIFF, save_projection_geometry=1)

    projection, geometry = utility.load_projection(temp_projection)

    mapped_values, uv_px = map_geometry_2_projection(
        source_position,
        detector_position,
        detctor_orientation,
        points,
        projection,
        cyl_proj.pixel_pitch_mm,
    )

    values_calc, angles_calc = map_source_2_cylinder(
        source_position,
        maps,
        points,
        cyl_proj.detector_radius_mm,
        cyl_proj.x_px,
        cyl_proj.x_resolution_mm,
    )

    for i in range(values_calc.shape[0]):
        print(
            f"Step {i}: Value Cylinder {values_calc[i]:2.2f} \t/\t Value Projection {mapped_values[i]:2.2f}"
        )

    fig = plt.figure(figsize=(24, 10))
    subfig1: figure.Figure
    subfig2: figure.Figure
    subfig1, subfig2 = fig.subfigures(1, 2)
    ax1 = subfig1.add_subplot(111)
    ax1.set_title("Projection Mapping")
    im = ax1.imshow(projection, vmin=maps.min(), vmax=maps.max())

    counter = 0
    number_of_maps = maps.shape[2]

    color = jnp.linspace(0, 1, 9)[:number_of_maps]
    cmap = plt.get_cmap("hsv")
    colors = cmap(color)
    ax1.scatter(uv_px[0, :9], uv_px[1, :9], c=colors)
    ax1.set_xlabel(r"$u$ / px")
    ax1.set_ylabel(r"$v$ / px")
    subfig1.colorbar(im, ax=ax1, orientation="vertical", label="Ray Length / mm")

    ax2 = subfig2.add_subplot(111)
    ax2.set_title("Cylindrical Mapping")
    ax2.axis(False)
    nums_subplots = 1
    for _ in range(nums_subplots):
        for _ in range(nums_subplots):
            if counter >= number_of_maps:
                continue
            ax = subfig2.add_subplot(nums_subplots, nums_subplots, counter + 1)
            ax.imshow(maps[:, :, counter], vmin=maps.min(), vmax=maps.max())
            ax.scatter(
                angles_calc[0, counter], angles_calc[1, counter], c=colors[counter]
            )
            ax.set_xlabel(r"$z$ / mm")
            ax.set_ylabel(r"$\alpha$ / rad")
            ax.set_xticks([0, 1000, 2000])
            ax.set_xticklabels([r"$-3141.5$", r"$0.$", r"$3141.5$"])
            ax.set_yticks([0, 1000, 2000])
            ax.set_yticklabels([r"$-\pi$", r"$0.$", r"$\pi$"])

            counter += 1

    plt.tight_layout()
    plt.savefig(TEMP_FOLDER / "mapping.png")


if __name__ == "__main__":
    main()
