from jax import numpy as jnp
import dm_pix as pix
from jax import jit
from jax.scipy.spatial.transform import Rotation


@jit
def map_source_2_cylinder(
    source: jnp.ndarray,
    maps: jnp.ndarray,
    map_positions: jnp.ndarray,
    radius: float = 1000.0,
    angle_discretisation: int = 2000,
    pitch: float = jnp.pi,
) -> tuple[jnp.ndarray, jnp.ndarray]:
    """
    source:        1 x 3
    maps:          u x v x m
    map_positions:     3 x m

    ---

    cylinder_value:        m x 1
    """

    # map_positions = map_positions.T # m x 3

    angles = cylindrical_angles(
        source, map_positions, radius, angle_discretisation, pitch
    )

    u = jnp.expand_dims(angles[0], 0)  # 1 x m
    v = jnp.expand_dims(angles[1], 0)  # 1 x m
    w = jnp.arange(0, maps.shape[2])[jnp.newaxis, ...]  # +0.0001 # 1 x m

    uvw = jnp.concatenate([v, u, w], 0)  # 3 x m
    values = pix.flat_nd_linear_interpolate(maps, uvw)

    return values, uvw  # m,


@jit
def cylindrical_angles(
    source: jnp.ndarray,
    map_positions: jnp.ndarray,
    radius: float = 1000.0,
    angle_discretisation: int = 2000,
    pitch: float = jnp.pi,
) -> tuple[jnp.ndarray, jnp.ndarray]:
    source = source.reshape((-1, 1))  # 1 x 3
    directions = map_positions - source  # m x 3
    direction_norm_factor = jnp.linalg.norm(directions[:2], axis=0, keepdims=True)
    directions = directions / direction_norm_factor  # m x 3
    intersection = directions * radius
    # sign = jnp.where(intersection[2]>=0, 1, -1)
    # intersection = intersection * sign

    z_value = intersection[2] / pitch
    z = z_value + (angle_discretisation) / 2.0
    beta_value = (
        (jnp.arctan2(intersection[0], intersection[1]) / jnp.pi)
        * angle_discretisation
        / 2
    )
    beta = beta_value - ((angle_discretisation - 1.0) / 4.0)
    angles = jnp.concatenate(
        (jnp.expand_dims(z, 0) % 2000, jnp.expand_dims(beta, 0) % 2000), axis=0
    )
    return angles


@jit
def map_geometry_2_projection(
    source: jnp.ndarray,
    detector: jnp.ndarray,
    detector_orientation: jnp.ndarray,
    map_positions: jnp.ndarray,
    projections: jnp.ndarray,
    pixel_pitch_mm: float = 0.139,
):
    """
    source:               1 x 3
    detector:             1 x 3
    detector_orientation: 1 x 4
    map_positions:            3 x m
    projections:                  u x v
    pixel_pitch_mm:               1,

    ---

    projection_value:             m
    """
    source = source.reshape((1, 3))
    detector = detector.reshape((1, 3))
    detector_orientation = detector_orientation.reshape((1, 4))

    projection_geometry = projection_matrix(
        source, detector, detector_orientation
    )  # 3 x 4
    map_positions_homogen = jnp.ones((1, map_positions.shape[1]))  # 1 x m
    map_positions_homogen = jnp.concatenate(
        (map_positions, map_positions_homogen), 0
    )  # 4 x m
    uv_px = projection_geometry @ map_positions_homogen  # 3 x m
    uv_px = uv_px / uv_px[2]
    uv_px = uv_px / pixel_pitch_mm
    uv_px = uv_px.at[0, :].multiply(-1)
    uv_px = uv_px.at[0, :].add((projections.shape[-2]) / 2.0)
    # uv_px = uv_px.at[0,:].multiply(-1)
    uv_px = uv_px.at[1, :].add((projections.shape[-2]) / 2.0)
    uv_px = uv_px[:2]

    uv_px_flip = jnp.flip(uv_px, 0)
    # uv_px_flip = uv_px

    return pix.flat_nd_linear_interpolate(projections, uv_px_flip), uv_px


@jit
def projection_matrix(
    source: jnp.ndarray, detector: jnp.ndarray, detector_orientation: jnp.ndarray
) -> jnp.ndarray:
    """
    source:               1 x 3
    detector:             1 x 3
    detector_orientation: 1 x 4

    ---

    projection_matrix =   3 x 4
    """
    detector_orientation = detector_orientation / detector_orientation[1, 3]
    rotation_matrix = Rotation.from_quat(detector_orientation).as_matrix()
    detector_horizontal_vector = -rotation_matrix[0, :, 0]
    detector_vertical_vector = -rotation_matrix[0, :, 1]
    p3x3 = jnp.vstack(
        [
            detector_horizontal_vector,
            detector_vertical_vector,
            (detector - source).reshape((-1, 3)),
        ]
    ).T
    p3x3_inv = jnp.linalg.inv(p3x3)
    p4 = p3x3_inv @ (-source).reshape((3, 1))
    matrix = jnp.concatenate([p3x3_inv, p4], 1)  # 3 x 4
    return matrix
