try:
    from artist_pythonlib import API, utility, SAVEMODES  # type: ignore
except ModuleNotFoundError:
    from warnings import warn

    warn(
        "The module `artistlib`is not installed. The simulation module is not 100\% usable! \nInstall: https://github.com/wittlsn/aRTist-PythonLib"
    )

    def API():
        return None

    utility = None
    SAVEMODES = None
from pathlib import Path
import numpy as np
import os
import importlib.resources
from time import sleep
from cyxtrax.common.mesh_object import MeshObject

# !!!!!!!!!!!!!!!!!!!
# Read carefully
#!!!!!!!!!!!!!!!!?!!!


# This Script assumes that the detector has a size of 6283.19 mm / 2 pi
# and the curvature is set to the y axis with center on the source.

# The source Position has to be on [0, 0, 0]
# The detector has to be on Position [1000,0, 0] and orientation [0, 90, 0] / curvature y

# The Script assumes this geometry and only moves all the objects!
# Preview must be reseted!!!

GLOBAL_COUNTER = 0


def set_cylindrical_mode(api: API):
    with importlib.resources.path("cyxtrax.data", "cylinder.aRTist") as file_path:
        print(f"Load: {file_path}")
        api.load_project(file_path)


def set_cone_mode(api: API):
    with importlib.resources.path("cyxtrax.data", "cone.aRTist") as file_path:
        print(f"Load: {file_path}")
        api.load_project(file_path)


def set_cone_poly_mode(api: API):
    with importlib.resources.path("cyxtrax.data", "cone_poly.aRTist") as file_path:
        print(f"Load: {file_path}")
        api.load_project(file_path)


class CylindricalProjection:
    def __init__(self, api: API = API()) -> None:
        self.api = api
        self.x_resolution_mm = np.pi
        self.y_resolution_mm = np.pi
        self.pixel_pitch_mm = 0.139
        self.x_px = 2000
        self.y_px = 2000
        self.detector_radius_mm = 1000.0
        self.objects = list()
        self._cylindrical = True

    @property
    def cylindrical_mode(self) -> bool:
        return self._cylindrical

    @property
    def cone_mode(self) -> bool:
        return not self._cylindrical

    def translate(self, position: np.ndarray):
        # for obj in self.objects:
        x_d, y_d, z_d = 1000, 0, 0
        x_s, y_s, z_s = 0, 0, 0
        x_p, y_p, z_p = position[0], position[1], position[2]

        self.api.rc.send(
            f'::PartList::Invoke {str("S")} SetPosition {str(x_s + x_p)} {str(y_s + y_p)} {str(z_s + z_p)};'
        )
        self.api.rc.send(
            f'::PartList::Invoke {str("S")} SetRefPos {str(x_s + x_p)} {str(y_s + y_p)} {str(z_s + z_p)};'
        )

        self.api.rc.send(
            f'::PartList::Invoke {str("D")} SetRefPos {str(x_s + x_p)} {str(y_s + y_p)} {str(z_s + z_p)};'
        )
        self.api.rc.send(
            f'::PartList::Invoke {str("D")} SetPosition {str(x_d + x_p)} {str(y_d + y_p)} {str(z_d + z_p)};'
        )
        self.api.rc.send("::XDetector::SetDownCurvedView;")

    def compute_projection(
        self,
        position: np.ndarray,
        temp_file_path: Path = Path(r"C:\data"),
        output_full_ray_projection: bool = True,
    ) -> np.ndarray:
        global GLOBAL_COUNTER
        self.translate(position)
        temp_file = temp_file_path / f"temp_{GLOBAL_COUNTER:01}.tiff"
        GLOBAL_COUNTER += 1
        GLOBAL_COUNTER = GLOBAL_COUNTER % 10

        self.api.save_image(
            temp_file, save_projection_geometry=False, save_mode=SAVEMODES.FLOAT_TIFF
        )

        while not temp_file.exists():
            print("sleepy sleep sleep ...")
            # The computation of the cylindrical xray tansforms takes time ... aRTist is not programmed for it.
            # This catches problems with the computation and the remote control interfaces
            # It is ugly ... but works ...
            # Also if the mesh is not waterprof the triangle filter of artist seems to have problems.

            sleep(2.0)

        image = utility.load_projection(temp_file, load_projection_geometry=False)[0]
        os.remove(temp_file)

        if output_full_ray_projection:
            return self.convert_rays(image)
        else:
            return image

    def convert_rays(self, image):
        inter_image = np.roll(image, 1000, 0)
        image += np.flip(inter_image, 1)
        return image

    def set_cylindrical_mode(self):
        set_cylindrical_mode(self.api)
        self.x_px = 2000.0
        self.y_px = 2000.0
        self.x_resolution_mm = np.pi
        self.y_resolution_mm = np.pi
        self.detector_radius_mm = 1000.0
        self._cylindrical = True
        print(
            "Care! The default pixel pitch etc is set. If changes are made change the state of this class ..."
        )

    def set_cone_mode(self):
        set_cone_mode(self.api)
        self.pixel_pitch_mm = 0.139
        self._cylindrical = False
        print(
            "Care! The default pixel pitch etc is set. If changes are made change the state of this class ..."
        )


def load_mesh(
    cylindrical_projection: CylindricalProjection,
    object_list: list[MeshObject],
    cylindrical_mode: bool = True,
) -> bool:
    if cylindrical_mode:
        cylindrical_projection.set_cylindrical_mode()
    else:
        cylindrical_projection.set_cone_mode()

    for mesh_object in object_list:
        object_id = cylindrical_projection.api.load_part(mesh_object.object_path)
        cylindrical_projection.api.translate(
            object_id,
            mesh_object.position_mm[0],
            mesh_object.position_mm[1],
            mesh_object.position_mm[2],
        )
        cylindrical_projection.api.rotate_from_quat(
            object_id, mesh_object.orientation_quat
        )

    return True
