from matplotlib import pyplot as plt, figure
from jax import numpy as jnp
from pathlib import Path


def save_map_plot_with_index(
    maps: jnp.ndarray, index: int, save_path: Path, figsize=(8, 8), title: str = None
):
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.imshow(maps[:, :, index])
    ax.set_xlabel(r"$z$ / mm")
    ax.set_ylabel(r"$\alpha$ / rad")
    ax.set_xticks([0, int(maps.shape[0] // 2), int(maps.shape[0])])
    ax.set_xticklabels([r"$-3141.5$", r"$0.$", r"$3141.5$"])
    ax.set_yticks([0, int(maps.shape[1] // 2), int(maps.shape[1])])
    ax.set_yticklabels([r"$-\pi$", r"$0.$", r"$\pi$"])

    if title is not None:
        ax.set_title(title)

    plt.tight_layout()

    plt.savefig(save_path)
    plt.close(fig)


def voxel_cone_maps(
    maps: jnp.ndarray,
    projection: jnp.ndarray,
    cone_coordinates: jnp.ndarray,
    cylindrical_coordinates: jnp.ndarray,
    save_path: Path,
    figsize=(24, 10),
    title: str = None,
    vmin: bool = False,
):
    fig = plt.figure(figsize=figsize)
    subfig1: figure.Figure
    subfig2: figure.Figure
    subfig1, subfig2 = fig.subfigures(1, 2)
    ax1 = subfig1.add_subplot(111)
    ax1.set_title("Projection Mapping")
    im = ax1.imshow(projection, vmin=maps.min(), vmax=maps.max())

    counter = 0
    number_of_maps = maps.shape[2]

    color = jnp.linspace(0, 1, 9)[:number_of_maps]
    cmap = plt.get_cmap("hsv")
    colors = cmap(color)

    ax1.set_xlabel(r"$u$ / px")
    ax1.set_ylabel(r"$v$ / px")
    subfig1.colorbar(im, ax=ax1, orientation="vertical", label="Ray Length / mm")

    ax2 = subfig2.add_subplot(111)
    ax2.set_title("Cylindrical Mapping")
    ax2.axis(False)

    side_length = int(jnp.round(number_of_maps ** (1.0 / 3.0)))

    for z in [0, side_length - 1]:
        for y in [0, side_length - 1]:
            for x in [0, side_length - 1]:
                position = x + y * side_length + z * side_length**2
                if counter >= number_of_maps:
                    continue
                ax = subfig2.add_subplot(3, 3, counter + 1)
                if vmin:
                    ax.imshow(maps[:, :, position].T, vmin=maps.min(), vmax=maps.max())
                else:
                    ax.imshow(maps[:, :, position].T, vmin=maps.min())
                ax.scatter(
                    cylindrical_coordinates[0, position],
                    cylindrical_coordinates[1, position],
                    c=colors[counter],
                )
                ax1.scatter(
                    cone_coordinates[0, position],
                    cone_coordinates[1, position],
                    c=colors[counter],
                )
                ax.set_ylabel(r"$z$ / mm")
                ax.set_xlabel(r"$\alpha$ / rad")
                ax.set_yticks([0, 1000, 2000])
                ax.set_yticklabels([r"$-3141.5$", r"$0.$", r"$3141.5$"])
                ax.set_xticks([0, 1000, 2000])
                ax.set_xticklabels([r"$-\pi$", r"$0.$", r"$\pi$"])

                counter += 1

    if title is not None:
        plt.title(title)

    plt.tight_layout()
    plt.savefig(save_path)


def voxel_cone_maps_values(
    maps: jnp.ndarray,
    projection: jnp.ndarray,
    cone_coordinates: jnp.ndarray,
    cylindrical_coordinates: jnp.ndarray,
    cylindrical_values: jnp.ndarray,
    save_path: Path,
    figsize=(24, 10),
    title: str = None,
):
    fig = plt.figure(figsize=figsize)
    subfig1: figure.Figure
    subfig2: figure.Figure
    subfig1, subfig2 = fig.subfigures(1, 2)
    ax1 = subfig1.add_subplot(111)
    ax1.set_title("Projection Mapping")
    im = ax1.imshow(projection, vmin=maps.min(), vmax=maps.max())

    counter = 0
    number_of_maps = maps.shape[2]

    color = jnp.linspace(0, 1, 9)[:number_of_maps]
    cmap = plt.get_cmap("hsv")
    colors = cmap(color)

    ax1.set_xlabel(r"$u$ / px")
    ax1.set_ylabel(r"$v$ / px")
    subfig1.colorbar(im, ax=ax1, orientation="vertical", label="Ray Length / mm")

    ax2 = subfig2.add_subplot(111)
    ax2.set_title("Cylindrical Mapping")
    ax2.axis(False)

    side_length = int(jnp.round(number_of_maps ** (1.0 / 3.0)))

    for z in [0, side_length - 1]:
        for y in [0, side_length - 1]:
            for x in [0, side_length - 1]:
                position = x + y * side_length + z * side_length**2
                if counter >= number_of_maps:
                    continue
                ax = subfig2.add_subplot(3, 3, counter + 1)
                ax.scatter(
                    cylindrical_values[0],
                    cylindrical_values[1],
                )
                ax.scatter(
                    cylindrical_coordinates[0, position],
                    cylindrical_coordinates[1, position],
                    c=colors[counter],
                )
                ax1.scatter(
                    cone_coordinates[0, position],
                    cone_coordinates[1, position],
                    c=colors[counter],
                )
                ax.set_ylabel(r"$z$ / mm")
                ax.set_xlabel(r"$\alpha$ / rad")
                ax.set_yticks([0, 1000, 2000])
                ax.set_yticklabels([r"$-3141.5$", r"$0.$", r"$3141.5$"])
                ax.set_xticks([0, 1000, 2000])
                ax.set_xticklabels([r"$-\pi$", r"$0.$", r"$\pi$"])

                counter += 1

    if title is not None:
        plt.title(title)

    plt.tight_layout()
    plt.savefig(save_path)


def voxel_maps(
    maps: jnp.ndarray,
    save_path: Path,
    figsize=(24, 10),
    title: str = None,
    vmin: bool = False,
):
    fig = plt.figure(figsize=figsize)

    ax2 = fig.add_subplot(111)
    ax2.set_title("Cylindrical Mapping")
    ax2.axis(False)

    number_of_maps = maps.shape[2]
    side_length = int(jnp.round(number_of_maps ** (1.0 / 3.0)))
    counter = 0
    for z in [0, side_length - 1]:
        for y in [0, side_length - 1]:
            for x in [0, side_length - 1]:
                position = x + y * side_length + z * side_length**2
                if counter >= number_of_maps:
                    continue
                ax = fig.add_subplot(3, 3, counter + 1)
                if vmin:
                    ax.imshow(maps[:, :, position].T, vmin=maps.min(), vmax=maps.max())
                else:
                    ax.imshow(maps[:, :, position].T, vmin=maps.min())
                ax.set_ylabel(r"$z$ / mm")
                ax.set_xlabel(r"$\alpha$ / rad")
                ax.set_yticks([0, 1000, 2000])
                ax.set_yticklabels([r"$-3141.5$", r"$0.$", r"$3141.5$"])
                ax.set_xticks([0, 1000, 2000])
                ax.set_xticklabels([r"$-\pi$", r"$0.$", r"$\pi$"])

                counter += 1

    if title is not None:
        plt.title(title)

    plt.tight_layout()
    plt.savefig(save_path)
